"""Labeled Property Graph (LPG) model tests.

Tests for LPG operations across different query languages:
- gql/ - GQL (ISO standard) - Primary
- cypher/ - Cypher compatibility
- gremlin/ - Gremlin traversals
- graphql/ - GraphQL queries
"""
